from lmnr.sdk.client.asynchronous.resources.agent import AsyncAgent
from lmnr.sdk.client.asynchronous.resources.browser_events import AsyncBrowserEvents
from lmnr.sdk.client.asynchronous.resources.evals import AsyncEvals
from lmnr.sdk.client.asynchronous.resources.tags import AsyncTags

__all__ = [
    "AsyncAgent",
    "AsyncEvals",
    "AsyncBrowserEvents",
    "AsyncTags",
]



"""Agent resource for interacting with Laminar agents."""

from typing import Generator, Literal, overload
import uuid

from lmnr.sdk.client.synchronous.resources.base import BaseResource
from opentelemetry import trace

from lmnr.sdk.types import (
    AgentOutput,
    LaminarSpanContext,
    ModelProvider,
    RunAgentRequest,
    RunAgentResponseChunk,
)


class Agent(BaseResource):
    """Resource for interacting with Laminar agents."""

    @overload
    def run(
        self,
        prompt: str,
        stream: Literal[True],
        parent_span_context: LaminarSpanContext | str | None = None,
        model_provider: ModelProvider | None = None,
        model: str | None = None,
        enable_thinking: bool = True,
        agent_state: str | None = None,
        storage_state: str | None = None,
        return_screenshots: bool = False,
        return_agent_state: bool = False,
        return_storage_state: bool = False,
        disable_give_control: bool = False,
        timeout: int | None = None,
        cdp_url: str | None = None,
        max_steps: int | None = None,
        thinking_token_budget: int | None = None,
        start_url: str | None = None,
        user_agent: str | None = None,
    ) -> Generator[RunAgentResponseChunk, None, None]:
        """Run Laminar index agent in streaming mode.

        Args:
            prompt (str): prompt for the agent
            stream (Literal[True]): whether to stream the agent's response
            parent_span_context (LaminarSpanContext | str | None, optional): span context if the agent is part of a trace
            model_provider (ModelProvider | None, optional): LLM model provider
            model (str | None, optional): LLM model name
            enable_thinking (bool, optional): whether to enable thinking on the underlying LLM. Default to True.
            agent_state (str | None, optional): the agent's state as returned by the previous agent run. Default to None.
            storage_state (str | None, optional): the browser's storage state as returned by the previous agent run. Default to None.
            return_screenshots (bool, optional): whether to return screenshots of the agent's states at every step. Default to False.
            return_agent_state (bool, optional): whether to return the agent's state. Default to False.
            return_storage_state (bool, optional): whether to return the storage state. Default to False.
            disable_give_control (bool, optional): whether to NOT give the agent additional direction to give control to the user for tasks such as login. Default to False.
            timeout (int | None, optional): timeout seconds for the agent's response. Default to None.
            cdp_url (str | None, optional): Chrome DevTools Protocol URL of an existing browser session. Default to None.
            max_steps (int | None, optional): maximum number of steps the agent can take. If not set, the backend will use a default value (currently 100). Default to None.
            thinking_token_budget (int | None, optional): maximum number of tokens the underlying LLM can spend on thinking in each step, if supported by the model. Default to None.
            start_url (str | None, optional): the URL to start the agent on. Must be a valid URL - refer to https://playwright.dev/docs/api/class-page#page-goto. If not specified, the agent infers this from the prompt. Default to None.
            user_agent (str | None, optional): the user to be sent to the browser. If not specified, Laminar uses the default user agent. Default to None.
        Returns:
            Generator[RunAgentResponseChunk, None, None]: a generator of response chunks
        """
        pass

    @overload
    def run(
        self,
        prompt: str,
        parent_span_context: LaminarSpanContext | str | None = None,
        model_provider: ModelProvider | None = None,
        model: str | None = None,
        enable_thinking: bool = True,
        agent_state: str | None = None,
        storage_state: str | None = None,
        return_screenshots: bool = False,
        return_agent_state: bool = False,
        disable_give_control: bool = False,
        return_storage_state: bool = False,
        timeout: int | None = None,
        cdp_url: str | None = None,
        max_steps: int | None = None,
        thinking_token_budget: int | None = None,
        start_url: str | None = None,
        user_agent: str | None = None,
    ) -> AgentOutput:
        """Run Laminar index agent.

        Args:
            prompt (str): prompt for the agent
            parent_span_context (LaminarSpanContext | str | None, optional): span context if the agent is part of a trace
            model_provider (ModelProvider | None, optional): LLM model provider
            model (str | None, optional): LLM model name
            enable_thinking (bool, optional): whether to enable thinking on the underlying LLM. Default to True.
            agent_state (str | None, optional): the agent's state as returned by the previous agent run. Default to None.
            storage_state (str | None, optional): the browser's storage state as returned by the previous agent run. Default to None.
            return_screenshots (bool, optional): whether to return screenshots of the agent's states at every step. Default to False.
            return_agent_state (bool, optional): whether to return the agent's state. Default to False.
            return_storage_state (bool, optional): whether to return the storage state. Default to False.
            disable_give_control (bool, optional): whether to NOT give the agent additional direction to give control to the user for tasks such as login. Default to False.
            timeout (int | None, optional): timeout seconds for the agent's response. Default to None.
            cdp_url (str | None, optional): Chrome DevTools Protocol URL of an existing browser session. Default to None.
            max_steps (int | None, optional): maximum number of steps the agent can take. If not set, the backend will use a default value (currently 100). Default to None.
            thinking_token_budget (int | None, optional): maximum number of tokens the underlying LLM can spend on thinking in each step, if supported by the model. Default to None.
            start_url (str | None, optional): the URL to start the agent on. Must be a valid URL - refer to https://playwright.dev/docs/api/class-page#page-goto. If not specified, the agent infers this from the prompt. Default to None.
            user_agent (str | None, optional): the user to be sent to the browser. If not specified, Laminar uses the default user agent. Default to None.

        Returns:
            AgentOutput: agent output
        """
        pass

    @overload
    def run(
        self,
        prompt: str,
        parent_span_context: LaminarSpanContext | str | None = None,
        model_provider: ModelProvider | None = None,
        model: str | None = None,
        stream: Literal[False] = False,
        enable_thinking: bool = True,
        agent_state: str | None = None,
        storage_state: str | None = None,
        return_screenshots: bool = False,
        return_agent_state: bool = False,
        return_storage_state: bool = False,
        disable_give_control: bool = False,
        timeout: int | None = None,
        cdp_url: str | None = None,
        max_steps: int | None = None,
        thinking_token_budget: int | None = None,
        start_url: str | None = None,
        user_agent: str | None = None,
    ) -> AgentOutput:
        """Run Laminar index agent.

        Args:
            prompt (str): prompt for the agent
            parent_span_context (LaminarSpanContext | str | None, optional): span context if the agent is part of a trace
            model_provider (ModelProvider | None, optional): LLM model provider
            model (str | None, optional): LLM model name
            stream (Literal[False], optional): whether to stream the agent's response
            enable_thinking (bool, optional): whether to enable thinking on the underlying LLM. Default to True.
            agent_state (str | None, optional): the agent's state as returned by the previous agent run. Default to None.
            storage_state (str | None, optional): the browser's storage state as returned by the previous agent run. Default to None.
            return_screenshots (bool, optional): whether to return screenshots of the agent's states at every step. Default to False.
            return_agent_state (bool, optional): whether to return the agent's state. Default to False.
            return_storage_state (bool, optional): whether to return the storage state. Default to False.
            disable_give_control (bool, optional): whether to NOT give the agent additional direction to give control to the user for tasks such as login. Default to False.
            timeout (int | None, optional): timeout seconds for the agent's response. Default to None.
            cdp_url (str | None, optional): Chrome DevTools Protocol URL of an existing browser session. Default to None.
            max_steps (int | None, optional): maximum number of steps the agent can take. If not set, the backend will use a default value (currently 100). Default to None.
            thinking_token_budget (int | None, optional): maximum number of tokens the underlying LLM can spend on thinking in each step, if supported by the model. Default to None.
            start_url (str | None, optional): the URL to start the agent on. Must be a valid URL - refer to https://playwright.dev/docs/api/class-page#page-goto. If not specified, the agent infers this from the prompt. Default to None.
            user_agent (str | None, optional): the user to be sent to the browser. If not specified, Laminar uses the default user agent. Default to None.

        Returns:
            AgentOutput: agent output
        """
        pass

    def run(
        self,
        prompt: str,
        parent_span_context: LaminarSpanContext | str | None = None,
        model_provider: ModelProvider | None = None,
        model: str | None = None,
        stream: bool = False,
        enable_thinking: bool = True,
        agent_state: str | None = None,
        storage_state: str | None = None,
        return_screenshots: bool = False,
        return_agent_state: bool = False,
        return_storage_state: bool = False,
        disable_give_control: bool = False,
        timeout: int | None = None,
        cdp_url: str | None = None,
        max_steps: int | None = None,
        thinking_token_budget: int | None = None,
        start_url: str | None = None,
        user_agent: str | None = None,
    ) -> AgentOutput | Generator[RunAgentResponseChunk, None, None]:
        """Run Laminar index agent.

        Args:
            prompt (str): prompt for the agent
            parent_span_context (LaminarSpanContext | str | None, optional): span context if the agent is part of a trace
            model_provider (ModelProvider | None, optional): LLM model provider
            model (str | None, optional): LLM model name
            stream (bool, optional): whether to stream the agent's response
            enable_thinking (bool, optional): whether to enable thinking on the underlying LLM. Default to True.
            agent_state (str | None, optional): the agent's state as returned by the previous agent run. Default to None.
            storage_state (str | None, optional): the browser's storage state as returned by the previous agent run. Default to None.
            return_screenshots (bool, optional): whether to return screenshots of the agent's states at every step. Default to False.
            return_agent_state (bool, optional): whether to return the agent's state. Default to False.
            return_storage_state (bool, optional): whether to return the storage state. Default to False.
            disable_give_control (bool, optional): whether to NOT give the agent additional direction to give control to the user for tasks such as login. Default to False.
            timeout (int | None, optional): timeout seconds for the agent's response. Default to None.
            cdp_url (str | None, optional): Chrome DevTools Protocol URL of an existing browser session. Default to None.
            max_steps (int | None, optional): maximum number of steps the agent can take. If not set, the backend will use a default value (currently 100). Default to None.
            thinking_token_budget (int | None, optional): maximum number of tokens the underlying LLM can spend on thinking in each step, if supported by the model. Default to None.
            start_url (str | None, optional): the URL to start the agent on. Must be a valid URL - refer to https://playwright.dev/docs/api/class-page#page-goto. If not specified, the agent infers this from the prompt. Default to None.
            user_agent (str | None, optional): the user to be sent to the browser. If not specified, Laminar uses the default user agent. Default to None.

        Returns:
            AgentOutput | Generator[RunAgentResponseChunk, None, None]: agent output or a generator of response chunks
        """
        if parent_span_context is None:
            span = trace.get_current_span()
            if span != trace.INVALID_SPAN:
                parent_span_context = LaminarSpanContext(
                    trace_id=uuid.UUID(int=span.get_span_context().trace_id),
                    span_id=uuid.UUID(int=span.get_span_context().span_id),
                    is_remote=span.get_span_context().is_remote,
                )
        if parent_span_context is not None and isinstance(
            parent_span_context, LaminarSpanContext
        ):
            parent_span_context = str(parent_span_context)
        request = RunAgentRequest(
            prompt=prompt,
            parent_span_context=parent_span_context,
            model_provider=model_provider,
            model=model,
            agent_state=agent_state,
            storage_state=storage_state,
            # We always connect to stream, because our network configuration
            # has a hard fixed idle timeout of 350 seconds.
            # This means that if we don't stream, the connection will be closed.
            # For now, we just return the content of the final chunk if `stream` is
            # `False`.
            stream=True,
            enable_thinking=enable_thinking,
            return_screenshots=return_screenshots,
            return_agent_state=return_agent_state,
            return_storage_state=return_storage_state,
            timeout=timeout,
            cdp_url=cdp_url,
            max_steps=max_steps,
            thinking_token_budget=thinking_token_budget,
            start_url=start_url,
            disable_give_control=disable_give_control,
            user_agent=user_agent,
        )

        # For streaming case, use a generator function
        if stream:
            return self.__run_streaming(request)
        else:
            # For non-streaming case, process all chunks and return the final result
            return self.__run_non_streaming(request)

    def __run_streaming(
        self, request: RunAgentRequest
    ) -> Generator[RunAgentResponseChunk, None, None]:
        """Run agent in streaming mode.

        Args:
            request (RunAgentRequest): The request to run the agent with.

        Yields:
            RunAgentResponseChunk: Chunks of the agent's response.
        """
        with self._client.stream(
            "POST",
            self._base_url + "/v1/agent/run",
            json=request.model_dump(by_alias=True),
            headers=self._headers(),
        ) as response:
            if response.status_code != 200:
                raise RuntimeError(response.read())
            for line in response.iter_lines():
                line = str(line)
                if line.startswith("[DONE]"):
                    break
                if not line.startswith("data: "):
                    continue
                line = line[6:]
                if line:
                    chunk = RunAgentResponseChunk.model_validate_json(line)
                    yield chunk.root
                    if chunk.root.chunk_type in ["finalOutput", "error", "timeout"]:
                        break

    def __run_non_streaming(self, request: RunAgentRequest) -> AgentOutput:
        """Run agent in non-streaming mode.

        Args:
            request (RunAgentRequest): The request to run the agent with.

        Returns:
            AgentOutput: The agent's output.
        """
        final_chunk = None

        with self._client.stream(
            "POST",
            self._base_url + "/v1/agent/run",
            json=request.model_dump(by_alias=True),
            headers=self._headers(),
        ) as response:
            if response.status_code != 200:
                raise RuntimeError(response.read())
            for line in response.iter_lines():
                line = str(line)
                if line.startswith("[DONE]"):
                    break
                if not line.startswith("data: "):
                    continue
                line = line[6:]
                if line:
                    chunk = RunAgentResponseChunk.model_validate_json(line)
                    if chunk.root.chunk_type == "finalOutput":
                        final_chunk = chunk.root
                    elif chunk.root.chunk_type == "error":
                        raise RuntimeError(chunk.root.error)
                    elif chunk.root.chunk_type == "timeout":
                        raise TimeoutError("Agent timed out")

        return final_chunk.content if final_chunk is not None else AgentOutput()



"""Evals resource for interacting with Laminar evaluations API."""

import uuid
import urllib.parse

from lmnr.sdk.client.synchronous.resources.base import BaseResource
from lmnr.sdk.types import (
    InitEvaluationResponse,
    EvaluationResultDatapoint,
    PartialEvaluationDatapoint,
    GetDatapointsResponse,
)


class Evals(BaseResource):
    """Resource for interacting with Laminar evaluations API."""

    def init(
        self, name: str | None = None, group_name: str | None = None
    ) -> InitEvaluationResponse:
        """Initialize a new evaluation.

        Args:
            name (str | None, optional): Name of the evaluation. Defaults to None.
            group_name (str | None, optional): Group name for the evaluation. Defaults to None.

        Returns:
            InitEvaluationResponse: The response from the initialization request.
        """
        response = self._client.post(
            self._base_url + "/v1/evals",
            json={
                "name": name,
                "groupName": group_name,
            },
            headers=self._headers(),
        )
        if response.status_code != 200:
            if response.status_code == 401:
                raise ValueError("Unauthorized. Please check your project API key.")
            raise ValueError(f"Error initializing evaluation: {response.text}")
        resp_json = response.json()
        return InitEvaluationResponse.model_validate(resp_json)

    def save_datapoints(
        self,
        eval_id: uuid.UUID,
        datapoints: list[EvaluationResultDatapoint | PartialEvaluationDatapoint],
        group_name: str | None = None,
    ):
        """Save evaluation datapoints.

        Args:
            eval_id (uuid.UUID): The evaluation ID.
            datapoints (list[EvaluationResultDatapoint | PartialEvaluationDatapoint]): The datapoints to save.
            group_name (str | None, optional): Group name for the datapoints. Defaults to None.

        Raises:
            ValueError: If there's an error saving the datapoints.
        """
        response = self._client.post(
            self._base_url + f"/v1/evals/{eval_id}/datapoints",
            json={
                "points": [datapoint.to_dict() for datapoint in datapoints],
                "groupName": group_name,
            },
            headers=self._headers(),
        )
        if response.status_code != 200:
            raise ValueError(f"Error saving evaluation datapoints: {response.text}")

    def get_datapoints(
        self,
        dataset_name: str,
        offset: int,
        limit: int,
    ) -> GetDatapointsResponse:
        """Get datapoints from a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            offset (int): The offset to start from.
            limit (int): The maximum number of datapoints to return.

        Returns:
            GetDatapointsResponse: The response containing the datapoints.

        Raises:
            ValueError: If there's an error fetching the datapoints.
        """
        params = {"name": dataset_name, "offset": offset, "limit": limit}
        url = (
            self._base_url + "/v1/datasets/datapoints?" + urllib.parse.urlencode(params)
        )
        response = self._client.get(url, headers=self._headers())
        if response.status_code != 200:
            try:
                resp_json = response.json()
                raise ValueError(
                    f"Error fetching datapoints: [{response.status_code}] {resp_json}"
                )
            except Exception:
                raise ValueError(
                    f"Error fetching datapoints: [{response.status_code}] {response.text}"
                )
        return GetDatapointsResponse.model_validate(response.json())



"""Resource for tagging traces."""

import json
import uuid

from lmnr.sdk.client.asynchronous.resources.base import BaseAsyncResource
from lmnr.sdk.log import get_default_logger

logger = get_default_logger(__name__)


class AsyncTags(BaseAsyncResource):
    """Resource for tagging traces."""

    async def tag(
        self,
        trace_id: str | int | uuid.UUID,
        tags: list[str] | str,
    ):
        """Tag a trace with a list of tags. Note that the trace must be ended
        before tagging it. You may want to call `Laminar.flush()` after the
        trace that you want to tag.

        Args:
            trace_id (str | int | uuid.UUID): The trace id to tag.
            tags (list[str] | str): The tag or list of tags to add to the trace.

        Raises:
            ValueError: If the trace id is not a valid UUID.

        Returns:
            list[dict]: The response from the server.

        Example:
        ```python
        from lmnr import Laminar, AsyncLaminarClient, observe

        Laminar.initialize()
        client = AsyncLaminarClient()
        trace_id = None

        @observe()
        def foo():
            trace_id = Laminar.get_trace_id()
            pass

        # make sure `foo` is called outside a trace context
        foo()

        # or make sure the trace is ended by this point
        Laminar.flush()

        await client.tags.tag(trace_id, "my_tag")
        ```
        """
        trace_tags = tags if isinstance(tags, list) else [tags]
        if isinstance(trace_id, uuid.UUID):
            trace_id = str(trace_id)
        elif isinstance(trace_id, int):
            trace_id = str(uuid.UUID(int=trace_id))
        elif isinstance(trace_id, str):
            uuid.UUID(trace_id)  # Will raise ValueError if invalid
        else:
            raise ValueError(f"Invalid trace id: {trace_id}")

        url = self._base_url + "/v1/tag"
        payload = {
            "traceId": trace_id,
            "names": trace_tags,
        }
        response = await self._client.post(
            url,
            content=json.dumps(payload),
            headers={
                **self._headers(),
            },
        )

        if response.status_code == 404:
            logger.warning(
                f"Trace {trace_id} not found. The trace may have not been ended yet."
            )
            return []

        if response.status_code != 200:
            raise ValueError(
                f"Failed to tag trace: [{response.status_code}] {response.text}"
            )
        return response.json()



import copy

from lmnr.opentelemetry_lib.tracing.attributes import (
    ASSOCIATION_PROPERTIES,
    TRACING_LEVEL,
)

from opentelemetry.context import Context, attach, set_value, get_value
from opentelemetry.sdk.trace import Span
from opentelemetry import trace


# TODO: delete this once deprecated Laminar.with_labels is removed. The logic
# should be moved into Laminar.set_tracing_level
def set_association_properties(properties: dict) -> None:
    attach(set_value("association_properties", properties))

    span = trace.get_current_span()
    _set_association_properties_attributes(span, properties)


# TODO: delete this once deprecated Laminar.with_labels is removed
def get_association_properties(context: Context | None = None) -> dict:
    return get_value("association_properties", context) or {}


# TODO: delete this once deprecated Laminar.with_labels is removed. The logic
# should be moved into Laminar.set_tracing_level
def update_association_properties(
    properties: dict,
    set_on_current_span: bool = True,
    context: Context | None = None,
) -> None:
    """Only adds or updates properties that are not already present"""
    association_properties = get_value("association_properties", context) or {}
    association_properties.update(properties)

    attach(set_value("association_properties", association_properties, context))

    if set_on_current_span:
        span = trace.get_current_span()
        _set_association_properties_attributes(span, properties)


# TODO: this logic should be moved into Laminar.set_tracing_level
def remove_association_properties(properties: dict) -> None:
    props: dict = copy.copy(get_value("association_properties") or {})
    for k in properties.keys():
        props.pop(k, None)
    set_association_properties(props)


def _set_association_properties_attributes(span: Span, properties: dict) -> None:
    if not span.is_recording():
        return
    for key, value in properties.items():
        if key == TRACING_LEVEL:
            span.set_attribute(f"lmnr.internal.{TRACING_LEVEL}", value)
            continue
        if (
            key in ["langgraph.edges", "langgraph.nodes"]
            and span.name != "LangGraph.workflow"
        ):
            continue
        span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{key}", value)



from lmnr.sdk.client.synchronous.resources.agent import Agent
from lmnr.sdk.client.synchronous.resources.browser_events import BrowserEvents
from lmnr.sdk.client.synchronous.resources.evals import Evals
from lmnr.sdk.client.synchronous.resources.tags import Tags

__all__ = ["Agent", "Evals", "BrowserEvents", "Tags"]



from contextlib import contextmanager
from contextvars import Context
import warnings
from typing_extensions import deprecated
from lmnr.opentelemetry_lib import TracerManager
from lmnr.opentelemetry_lib.tracing.instruments import Instruments
from lmnr.opentelemetry_lib.tracing.tracer import get_tracer
from lmnr.opentelemetry_lib.tracing.attributes import (
    ASSOCIATION_PROPERTIES,
    USER_ID,
    Attributes,
    SPAN_TYPE,
)
from lmnr.opentelemetry_lib import MAX_MANUAL_SPAN_PAYLOAD_SIZE
from lmnr.opentelemetry_lib.decorators import json_dumps
from opentelemetry import context as context_api, trace
from opentelemetry.context import attach, detach
from opentelemetry.trace import INVALID_TRACE_ID
from opentelemetry.sdk.trace.id_generator import RandomIdGenerator
from opentelemetry.util.types import AttributeValue

from typing import Any, Literal

import datetime
import logging
import os
import re
import uuid

from lmnr.opentelemetry_lib.tracing.attributes import (
    SESSION_ID,
    SPAN_INPUT,
    SPAN_OUTPUT,
    TRACE_TYPE,
)
from lmnr.opentelemetry_lib.tracing.context_properties import (
    get_association_properties,
    remove_association_properties,
    set_association_properties,
    update_association_properties,
)
from lmnr.sdk.utils import from_env, is_otel_attribute_value_type

from .log import VerboseColorfulFormatter

from .types import (
    LaminarSpanContext,
    TraceType,
    TracingLevel,
)


class Laminar:
    __project_api_key: str | None = None
    __initialized: bool = False
    __base_http_url: str | None = None

    @classmethod
    def initialize(
        cls,
        project_api_key: str | None = None,
        base_url: str | None = None,
        http_port: int | None = None,
        grpc_port: int | None = None,
        instruments: set[Instruments] | None = None,
        disabled_instruments: set[Instruments] | None = None,
        disable_batch: bool = False,
        max_export_batch_size: int | None = None,
        export_timeout_seconds: int | None = None,
        set_global_tracer_provider: bool = True,
        otel_logger_level: int = logging.ERROR,
    ):
        """Initialize Laminar context across the application.
        This method must be called before using any other Laminar methods or
        decorators.

        Args:
            project_api_key (str | None, optional): Laminar project api key.\
                            You can generate one by going to the projects\
                            settings page on the Laminar dashboard.\
                            If not specified, it will try to read from the\
                            LMNR_PROJECT_API_KEY environment variable\
                            in os.environ or in .env file.
                            Defaults to None.
            base_url (str | None, optional): Laminar API url. Do NOT include\
                            the port number, use `http_port` and `grpc_port`.\
                            If not specified, defaults to https://api.lmnr.ai.
            http_port (int | None, optional): Laminar API http port.\
                            If not specified, defaults to 443.
            grpc_port (int | None, optional): Laminar API grpc port.\
                            If not specified, defaults to 8443.
            instruments (set[Instruments] | None, optional): Instruments to\
                        enable. Defaults to all instruments. You can pass\
                        an empty set to disable all instruments. Read more:\
                        https://docs.lmnr.ai/tracing/automatic-instrumentation
            disabled_instruments (set[Instruments] | None, optional): Instruments to\
                        disable. Defaults to None.
            disable_batch (bool, optional): If set to True, spans will be sent\
                        immediately to the backend. Useful for debugging, but\
                        may cause performance overhead in production.
                        Defaults to False.
            export_timeout_seconds (int | None, optional): Timeout for the OTLP\
                        exporter. Defaults to 30 seconds (unlike the\
                        OpenTelemetry default of 10 seconds).
                        Defaults to None.
            set_global_tracer_provider (bool, optional): If set to True, the\
                        Laminar tracer provider will be set as the global\
                        tracer provider. OpenTelemetry allows only one tracer\
                        provider per app, so set this to False, if you are using\
                        another tracing library. Setting this to False may break\
                        some external instrumentations, e.g. LiteLLM.
                        Defaults to True.

        Raises:
            ValueError: If project API key is not set
        """
        if cls.is_initialized():
            cls.__logger.info(
                "Laminar is already initialized. Skipping initialization."
            )
            return

        cls.__project_api_key = project_api_key or from_env("LMNR_PROJECT_API_KEY")
        if not cls.__project_api_key:
            raise ValueError(
                "Please initialize the Laminar object with"
                " your project API key or set the LMNR_PROJECT_API_KEY"
                " environment variable in your environment or .env file"
            )

        cls._initialize_logger()

        url = base_url or from_env("LMNR_BASE_URL") or "https://api.lmnr.ai"
        url = url.rstrip("/")
        if not url.startswith("http"):
            url = f"https://{url}"
        if match := re.search(r":(\d{1,5})$", url):
            url = url[: -len(match.group(0))]
            if http_port is None:
                cls.__logger.info(f"Using HTTP port from base URL: {match.group(1)}")
                http_port = int(match.group(1))
            else:
                cls.__logger.info(f"Using HTTP port passed as an argument: {http_port}")

        cls.__initialized = True
        cls.__base_http_url = f"{url}:{http_port or 443}"

        if not os.getenv("OTEL_ATTRIBUTE_COUNT_LIMIT"):
            # each message is at least 2 attributes: role and content,
            # but the default attribute limit is 128, so raise it
            os.environ["OTEL_ATTRIBUTE_COUNT_LIMIT"] = "10000"

        TracerManager.init(
            base_url=url,
            http_port=http_port or 443,
            port=grpc_port or 8443,
            project_api_key=cls.__project_api_key,
            instruments=instruments,
            block_instruments=disabled_instruments,
            disable_batch=disable_batch,
            max_export_batch_size=max_export_batch_size,
            timeout_seconds=export_timeout_seconds,
            set_global_tracer_provider=set_global_tracer_provider,
            otel_logger_level=otel_logger_level,
        )

    @classmethod
    def is_initialized(cls):
        """Check if Laminar is initialized. A utility to make sure other
        methods are called after initialization.

        Returns:
            bool: True if Laminar is initialized, False otherwise
        """
        return cls.__initialized

    @classmethod
    def _initialize_logger(cls):
        cls.__logger = logging.getLogger(__name__)
        console_log_handler = logging.StreamHandler()
        console_log_handler.setFormatter(VerboseColorfulFormatter())
        cls.__logger.addHandler(console_log_handler)

    @classmethod
    def event(
        cls,
        name: str,
        value: AttributeValue | None = None,
        timestamp: datetime.datetime | int | None = None,
    ):
        """Associate an event with the current span. If using manual\
        instrumentation, use raw OpenTelemetry `span.add_event()` instead.\
       `value` will be saved as a `lmnr.event.value` attribute.

        Args:
            name (str): event name
            value (AttributeValue | None, optional): event value. Must be a\
                primitive type. Boolean `True` is assumed in the backend if\
                `value` is None.
                Defaults to None.
            timestamp (datetime.datetime | int | None, optional): If int, must\
                be epoch nanoseconds. If not specified, relies on the underlying\
                OpenTelemetry implementation. Defaults to None.
        """
        if timestamp and isinstance(timestamp, datetime.datetime):
            timestamp = int(timestamp.timestamp() * 1e9)

        event = {
            "lmnr.event.type": "default",
        }
        if value is not None:
            event["lmnr.event.value"] = value

        current_span = trace.get_current_span()
        if current_span == trace.INVALID_SPAN:
            cls.__logger.warning(
                "`Laminar().event()` called outside of span context. "
                f"Event '{name}' will not be recorded in the trace. "
                "Make sure to annotate the function with a decorator"
            )
            return

        current_span.add_event(name, event, timestamp)

    @classmethod
    @contextmanager
    def start_as_current_span(
        cls,
        name: str,
        input: Any = None,
        span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
        context: Context | None = None,
        labels: list[str] | None = None,
        parent_span_context: LaminarSpanContext | None = None,
        tags: list[str] | None = None,
    ):
        """Start a new span as the current span. Useful for manual
        instrumentation. If `span_type` is set to `"LLM"`, you should report
        usage and response attributes manually. See `Laminar.set_span_attributes`
        for more information.

        Usage example:
        ```python
        with Laminar.start_as_current_span("my_span", input="my_input") as span:
            await my_async_function()
            Laminar.set_span_output("my_output")`
        ```

        Args:
            name (str): name of the span
            input (Any, optional): input to the span. Will be sent as an\
                attribute, so must be json serializable. Defaults to None.
            span_type (Literal["DEFAULT", "LLM", "TOOL"], optional):\
                type of the span. If you use `"LLM"`, you should report usage\
                and response attributes manually. Defaults to "DEFAULT".
            context (Context | None, optional): raw OpenTelemetry context\
                to attach the span to. Defaults to None.
            parent_span_context (LaminarSpanContext | None, optional): parent\
                span context to use for the span. Useful for continuing traces\
                across services. If parent_span_context is a\
                raw OpenTelemetry span context, or if it is a dictionary or string\
                obtained from `Laminar.get_laminar_span_context_dict()` or\
                `Laminar.get_laminar_span_context_str()` respectively, it will be\
                converted to a `LaminarSpanContext` if possible. See also\
                `Laminar.get_span_context`, `Laminar.get_span_context_dict` and\
                `Laminar.get_span_context_str` for more information.
                Defaults to None.
            labels (list[str] | None, optional): [DEPRECATED] Use tags\
                instead. Labels to set for the span. Defaults to None.
            tags (list[str] | None, optional): tags to set for the span.
                Defaults to None.
        """

        if not cls.is_initialized():
            yield trace.NonRecordingSpan(
                trace.SpanContext(
                    trace_id=RandomIdGenerator().generate_trace_id(),
                    span_id=RandomIdGenerator().generate_span_id(),
                    is_remote=False,
                )
            )
            return

        with get_tracer() as tracer:
            ctx = context or context_api.get_current()
            if parent_span_context is not None:
                span_context = LaminarSpanContext.try_to_otel_span_context(
                    parent_span_context, cls.__logger
                )
                ctx = trace.set_span_in_context(
                    trace.NonRecordingSpan(span_context), ctx
                )
            ctx_token = attach(ctx)
            label_props = {}
            try:
                if labels:
                    warnings.warn(
                        "`Laminar.start_as_current_span` `labels` is deprecated. Use `tags` instead.",
                        DeprecationWarning,
                    )
                    label_props = {f"{ASSOCIATION_PROPERTIES}.labels": labels}
            except Exception:
                cls.__logger.warning(
                    f"`start_as_current_span` Could not set labels: {labels}. "
                    "They will be propagated to the next span."
                )
            tag_props = {}
            if tags:
                if isinstance(tags, list) and all(isinstance(tag, str) for tag in tags):
                    tag_props = {f"{ASSOCIATION_PROPERTIES}.tags": tags}
                else:
                    cls.__logger.warning(
                        f"`start_as_current_span` Could not set tags: {tags}. Tags must be a list of strings. "
                        "Tags will be ignored."
                    )
            with tracer.start_as_current_span(
                name,
                context=ctx,
                attributes={
                    SPAN_TYPE: span_type,
                    **(label_props),
                    **(tag_props),
                },
            ) as span:
                if input is not None:
                    serialized_input = json_dumps(input)
                    if len(serialized_input) > MAX_MANUAL_SPAN_PAYLOAD_SIZE:
                        span.set_attribute(
                            SPAN_INPUT,
                            "Laminar: input too large to record",
                        )
                    else:
                        span.set_attribute(
                            SPAN_INPUT,
                            serialized_input,
                        )
                yield span

            # TODO: Figure out if this is necessary
            try:
                detach(ctx_token)
            except Exception:
                pass

    @classmethod
    @contextmanager
    @deprecated(
        "Use `Laminar.set_span_tags` or the `tags` argument of "
        "`Laminar.start_as_current_span` or `Laminar.start_span` instead"
    )
    def with_labels(cls, labels: list[str], context: Context | None = None):
        """Set labels for spans within this `with` context. This is useful for
        adding labels to the spans created in the auto-instrumentations.

        Requirements:
        - Labels must be created in your project in advance.
        - Keys must be strings from your label names.
        - Values must be strings matching the label's allowed values.

        Usage example:
        ```python
        with Laminar.with_labels({"sentiment": "positive"}):
            openai_client.chat.completions.create()
        ```
        """
        warnings.warn(
            "`Laminar.with_labels` is deprecated. Use `Laminar.set_span_tags` or the `tags` argument of "
            "`Laminar.start_as_current_span` or `Laminar.start_span` instead",
            DeprecationWarning,
        )
        if not cls.is_initialized():
            yield
            return

        with get_tracer():
            label_props = labels.copy()
            prev_labels = get_association_properties(context).get("labels", [])
            update_association_properties(
                {"labels": prev_labels + label_props},
                set_on_current_span=False,
                context=context,
            )
            yield
            try:
                set_association_properties({"labels": prev_labels})
            except Exception:
                cls.__logger.warning(
                    f"`with_labels` Could not remove labels: {labels}. They will be "
                    "propagated to the next span."
                )
                pass

    @classmethod
    def start_span(
        cls,
        name: str,
        input: Any = None,
        span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
        context: Context | None = None,
        parent_span_context: LaminarSpanContext | None = None,
        labels: dict[str, str] | None = None,
        tags: list[str] | None = None,
    ):
        """Start a new span. Useful for manual instrumentation.
        If `span_type` is set to `"LLM"`, you should report usage and response
        attributes manually. See `Laminar.set_span_attributes` for more
        information.

        Usage example:
        ```python
        from src.lmnr import Laminar, use_span
        def foo(span):
            with use_span(span):
                with Laminar.start_as_current_span("foo_inner"):
                    some_function()
        
        def bar():
            with use_span(span):
                openai_client.chat.completions.create()
        
        span = Laminar.start_span("outer")
        foo(span)
        bar(span)
        # IMPORTANT: End the span manually
        span.end()
        
        # Results in:
        # | outer
        # |   | foo
        # |   |   | foo_inner
        # |   | bar
        # |   |   | openai.chat
        ```

        Args:
            name (str): name of the span
            input (Any, optional): input to the span. Will be sent as an\
                attribute, so must be json serializable. Defaults to None.
            span_type (Literal["DEFAULT", "LLM", "TOOL"], optional):\
                type of the span. If you use `"LLM"`, you should report usage\
                and response attributes manually. Defaults to "DEFAULT".
            context (Context | None, optional): raw OpenTelemetry context\
                to attach the span to. Defaults to None.
            parent_span_context (LaminarSpanContext | None, optional): parent\
                span context to use for the span. Useful for continuing traces\
                across services. If parent_span_context is a\
                raw OpenTelemetry span context, or if it is a dictionary or string\
                obtained from `Laminar.get_laminar_span_context_dict()` or\
                `Laminar.get_laminar_span_context_str()` respectively, it will be\
                converted to a `LaminarSpanContext` if possible. See also\
                `Laminar.get_span_context`, `Laminar.get_span_context_dict` and\
                `Laminar.get_span_context_str` for more information.
                Defaults to None.
            tags (list[str] | None, optional): tags to set for the span.
                Defaults to None.
            labels (dict[str, str] | None, optional): [DEPRECATED] Use tags\
                instead. Labels to set for the span. Defaults to None.
        """
        if not cls.is_initialized():
            return trace.NonRecordingSpan(
                trace.SpanContext(
                    trace_id=RandomIdGenerator().generate_trace_id(),
                    span_id=RandomIdGenerator().generate_span_id(),
                    is_remote=False,
                )
            )

        with get_tracer() as tracer:
            ctx = context or context_api.get_current()
            if parent_span_context is not None:
                span_context = LaminarSpanContext.try_to_otel_span_context(
                    parent_span_context, cls.__logger
                )
                ctx = trace.set_span_in_context(
                    trace.NonRecordingSpan(span_context), ctx
                )
            label_props = {}
            try:
                if labels:
                    warnings.warn(
                        "`Laminar.start_span` `labels` is deprecated. Use `tags` instead.",
                        DeprecationWarning,
                    )
                    label_props = {
                        f"{ASSOCIATION_PROPERTIES}.labels": json_dumps(labels)
                    }
            except Exception:
                cls.__logger.warning(
                    f"`start_span` Could not set labels: {labels}. They will be "
                    "propagated to the next span."
                )
            tag_props = {}
            if tags:
                if isinstance(tags, list) and all(isinstance(tag, str) for tag in tags):
                    tag_props = {f"{ASSOCIATION_PROPERTIES}.tags": tags}
                else:
                    cls.__logger.warning(
                        f"`start_span` Could not set tags: {tags}. Tags must be a list of strings. "
                        + "Tags will be ignored."
                    )
            span = tracer.start_span(
                name,
                context=ctx,
                attributes={
                    SPAN_TYPE: span_type,
                    **(label_props),
                    **(tag_props),
                },
            )
            if input is not None:
                serialized_input = json_dumps(input)
                if len(serialized_input) > MAX_MANUAL_SPAN_PAYLOAD_SIZE:
                    span.set_attribute(
                        SPAN_INPUT,
                        "Laminar: input too large to record",
                    )
                else:
                    span.set_attribute(
                        SPAN_INPUT,
                        serialized_input,
                    )
            return span

    @classmethod
    def set_span_output(cls, output: Any = None):
        """Set the output of the current span. Useful for manual
        instrumentation.

        Args:
            output (Any, optional): output of the span. Will be sent as an\
                attribute, so must be json serializable. Defaults to None.
        """
        span = trace.get_current_span()
        if output is not None and span != trace.INVALID_SPAN:
            serialized_output = json_dumps(output)
            if len(serialized_output) > MAX_MANUAL_SPAN_PAYLOAD_SIZE:
                span.set_attribute(
                    SPAN_OUTPUT,
                    "Laminar: output too large to record",
                )
            else:
                span.set_attribute(SPAN_OUTPUT, serialized_output)

    @classmethod
    @contextmanager
    def set_tracing_level(self, level: TracingLevel):
        """Set the tracing level for the current span and the context
        (i.e. any children spans created from the current span in the current
        thread).

        Tracing level can be one of:
        - `TracingLevel.ALL`: Enable tracing for the current span and all
            children spans.
        - `TracingLevel.META_ONLY`: Enable tracing for the current span and all
            children spans, but only record metadata, e.g. tokens, costs.
        - `TracingLevel.OFF`: Disable recording any spans.

        Example:
        ```python
        from lmnr import Laminar, TracingLevel

        with Laminar.set_tracing_level(TracingLevel.META_ONLY):
            openai_client.chat.completions.create()
        ```
        """
        if level == TracingLevel.ALL:
            yield
        else:
            level = "meta_only" if level == TracingLevel.META_ONLY else "off"
            update_association_properties({"tracing_level": level})
            yield
            try:
                remove_association_properties({"tracing_level": level})
            except Exception:
                pass

    @classmethod
    def set_span_attributes(
        cls,
        attributes: dict[Attributes, Any],
    ):
        """Set attributes for the current span. Useful for manual
        instrumentation.
        Example:
        ```python
        with L.start_as_current_span(
            name="my_span_name", input=input["messages"], span_type="LLM"
        ):
            response = await my_custom_call_to_openai(input)
            L.set_span_output(response["choices"][0]["message"]["content"])
            L.set_span_attributes({
                Attributes.PROVIDER: 'openai',
                Attributes.REQUEST_MODEL: input["model"],
                Attributes.RESPONSE_MODEL: response["model"],
                Attributes.INPUT_TOKEN_COUNT: response["usage"]["prompt_tokens"],
                Attributes.OUTPUT_TOKEN_COUNT: response["usage"]["completion_tokens"],
            })
            # ...
        ```

        Args:
            attributes (dict[ATTRIBUTES, Any]): attributes to set for the span
        """
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            return

        for key, value in attributes.items():
            # Python 3.12+ should do: if key not in Attributes:
            try:
                Attributes(key.value)
            except (TypeError, AttributeError):
                cls.__logger.warning(
                    f"Attribute {key} is not a valid Laminar attribute."
                )
            if not isinstance(value, (str, int, float, bool)):
                span.set_attribute(key.value, json_dumps(value))
            else:
                span.set_attribute(key.value, value)

    @classmethod
    def get_laminar_span_context(
        cls, span: trace.Span | None = None
    ) -> LaminarSpanContext | None:
        """Get the laminar span context for a given span.
        If no span is provided, the current active span will be used.
        """
        span = span or trace.get_current_span()
        if span == trace.INVALID_SPAN:
            return None
        return LaminarSpanContext(
            trace_id=uuid.UUID(int=span.get_span_context().trace_id),
            span_id=uuid.UUID(int=span.get_span_context().span_id),
            is_remote=span.get_span_context().is_remote,
        )

    @classmethod
    def get_laminar_span_context_dict(
        cls, span: trace.Span | None = None
    ) -> dict | None:
        span_context = cls.get_laminar_span_context(span)
        if span_context is None:
            return None
        return span_context.model_dump()

    @classmethod
    def serialize_span_context(cls, span: trace.Span | None = None) -> str | None:
        """Get the laminar span context for a given span as a string.
        If no span is provided, the current active span will be used.

        This is useful for continuing a trace across services.

        Example:
        ```python
        # service A:
        with Laminar.start_as_current_span("service_a"):
            span_context = Laminar.serialize_span_context()
            # send span_context to service B
            call_service_b(request, headers={"laminar-span-context": span_context})

        # service B:
        def call_service_b(request, headers):
            span_context = Laminar.deserialize_span_context(headers["laminar-span-context"])
            with Laminar.start_as_current_span("service_b", parent_span_context=span_context):
                # rest of the function
                pass
        ```

        This will result in a trace like:
        ```
        service_a
          service_b
        ```
        """
        span_context = cls.get_laminar_span_context(span)
        if span_context is None:
            return None
        return str(span_context)

    @classmethod
    def deserialize_span_context(cls, span_context: dict | str) -> LaminarSpanContext:
        return LaminarSpanContext.deserialize(span_context)

    @classmethod
    def flush(cls) -> bool:
        """Flush the internal tracer.

        Returns:
            bool: True if the tracer was flushed, False otherwise
            (e.g. no tracer or timeout).
        """
        if not cls.is_initialized():
            return False
        return TracerManager.flush()

    @classmethod
    def shutdown(cls):
        if cls.is_initialized():
            TracerManager.shutdown()
            cls.__initialized = False

    @classmethod
    def set_span_tags(cls, tags: list[str]):
        """Set the tags for the current span.

        Args:
            tags (list[str]): Tags to set for the span.
        """
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set tags on")
            return
        if not isinstance(tags, list) or not all(isinstance(tag, str) for tag in tags):
            cls.__logger.warning(
                "Tags must be a list of strings. Tags will be ignored."
            )
            return
        span.set_attribute(f"{ASSOCIATION_PROPERTIES}.tags", tags)

    @classmethod
    @deprecated("Use `Laminar.set_trace_session_id` instead")
    def set_session(
        cls,
        session_id: str | None = None,
    ):
        """Set the session id for the current span and the context
        (i.e. any children spans created from the current span in the current
        thread).

        Args:
            session_id (str | None, optional): Custom session id.\
                            Useful to debug and group long-running\
                            sessions/conversations.
                            Defaults to None.
        """
        warnings.warn(
            "`Laminar.set_session` is deprecated. Use `Laminar.set_trace_session_id` instead",
            DeprecationWarning,
        )
        association_properties = {}
        if session_id is not None:
            association_properties[SESSION_ID] = session_id
        # update_association_properties(association_properties)
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set session id on")
            return
        if session_id is not None:
            span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{SESSION_ID}", session_id)

    @classmethod
    def set_trace_session_id(cls, session_id: str | None = None):
        """Set the session id for the current trace.
        Overrides any existing session id.

        Args:
            session_id (str | None, optional): Custom session id. Defaults to None.
        """
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set session id on")
            return
        if session_id is not None:
            span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{SESSION_ID}", session_id)

    @classmethod
    def set_trace_user_id(cls, user_id: str | None = None):
        """Set the user id for the current trace.
        Overrides any existing user id.

        Args:
            user_id (str | None, optional): Custom user id. Defaults to None.
        """
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set user id on")
            return
        if user_id is not None:
            span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{USER_ID}", user_id)

    @classmethod
    @deprecated("Use `Laminar.set_trace_metadata` instead")
    def set_metadata(cls, metadata: dict[str, str]):
        """Set the metadata for the current trace.

        Args:
            metadata (dict[str, str]): Metadata to set for the trace. Will be\
                sent as attributes, so must be json serializable.
        """
        warnings.warn(
            "`Laminar.set_metadata` is deprecated. Use `Laminar.set_trace_metadata` instead",
            DeprecationWarning,
        )
        props = {f"metadata.{k}": json_dumps(v) for k, v in metadata.items()}
        # update_association_properties(props)
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set metadata on")
            return
        for key, value in props.items():
            span.set_attribute(key, value)

    @classmethod
    def set_trace_metadata(cls, metadata: dict[str, AttributeValue]):
        """Set the metadata for the current trace.

        Args:
            metadata (dict[str, AttributeValue]): Metadata to set for the trace.
        """
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set metadata on")
            return
        for key, value in metadata.items():
            if is_otel_attribute_value_type(value):
                span.set_attribute(f"{ASSOCIATION_PROPERTIES}.metadata.{key}", value)
            else:
                span.set_attribute(
                    f"{ASSOCIATION_PROPERTIES}.metadata.{key}", json_dumps(value)
                )

    @classmethod
    def get_base_http_url(cls):
        return cls.__base_http_url

    @classmethod
    def get_project_api_key(cls):
        return cls.__project_api_key

    @classmethod
    def get_trace_id(cls) -> uuid.UUID | None:
        """Get the trace id for the current active span represented as a UUID.
        Returns None if there is no active span.

        Returns:
            uuid.UUID | None: The trace id for the current span, or None if\
            there is no active span.
        """
        trace_id = trace.get_current_span().get_span_context().trace_id
        if trace_id == INVALID_TRACE_ID:
            return None
        return uuid.UUID(int=trace_id)

    @classmethod
    def _headers(cls):
        assert cls.__project_api_key is not None, "Project API key is not set"
        return {
            "Authorization": "Bearer " + cls.__project_api_key,
            "Content-Type": "application/json",
        }

    @classmethod
    def _set_trace_type(
        cls,
        trace_type: TraceType,
    ):
        """Set the trace_type for the current span and the context
        Args:
            trace_type (TraceType): Type of the trace
        """
        span = trace.get_current_span()
        if span == trace.INVALID_SPAN:
            cls.__logger.warning("No active span to set trace type on")
            return
        span.set_attribute(f"{ASSOCIATION_PROPERTIES}.{TRACE_TYPE}", trace_type.value)



import atexit
import logging
import threading

from lmnr.opentelemetry_lib.tracing.processor import LaminarSpanProcessor
from lmnr.sdk.client.asynchronous.async_client import AsyncLaminarClient
from lmnr.sdk.client.synchronous.sync_client import LaminarClient
from lmnr.sdk.log import VerboseColorfulFormatter
from lmnr.opentelemetry_lib.tracing.instruments import (
    Instruments,
    init_instrumentations,
)

from opentelemetry import trace
from opentelemetry.instrumentation.threading import ThreadingInstrumentor
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider, SpanProcessor
from opentelemetry.sdk.trace.export import SpanExporter

TRACER_NAME = "lmnr.tracer"

MAX_EVENTS_OR_ATTRIBUTES_PER_SPAN = 5000


class TracerWrapper(object):
    resource_attributes: dict = {}
    enable_content_tracing: bool = True
    _lock = threading.Lock()
    _tracer_provider: TracerProvider | None = None
    _logger: logging.Logger
    _client: LaminarClient
    _async_client: AsyncLaminarClient
    _resource: Resource
    _span_processor: SpanProcessor

    def __new__(
        cls,
        disable_batch=False,
        exporter: SpanExporter | None = None,
        instruments: set[Instruments] | None = None,
        block_instruments: set[Instruments] | None = None,
        base_url: str = "https://api.lmnr.ai",
        port: int = 8443,
        http_port: int = 443,
        project_api_key: str | None = None,
        max_export_batch_size: int | None = None,
        force_http: bool = False,
        timeout_seconds: int = 10,
        set_global_tracer_provider: bool = True,
        otel_logger_level: int = logging.ERROR,
    ) -> "TracerWrapper":
        # Silence some opentelemetry warnings
        logging.getLogger("opentelemetry.trace").setLevel(otel_logger_level)

        base_http_url = f"{base_url}:{http_port}"
        with cls._lock:
            if not hasattr(cls, "instance"):
                cls._initialize_logger(cls)
                obj = super(TracerWrapper, cls).__new__(cls)

                obj._client = LaminarClient(
                    base_url=base_http_url,
                    project_api_key=project_api_key,
                )
                obj._async_client = AsyncLaminarClient(
                    base_url=base_http_url,
                    project_api_key=project_api_key,
                )

                obj._resource = Resource(attributes=TracerWrapper.resource_attributes)

                obj._span_processor = LaminarSpanProcessor(
                    base_url=base_url,
                    api_key=project_api_key,
                    port=http_port if force_http else port,
                    exporter=exporter,
                    max_export_batch_size=max_export_batch_size,
                    timeout_seconds=timeout_seconds,
                    force_http=force_http,
                    disable_batch=disable_batch,
                )

                lmnr_provider = TracerProvider(resource=obj._resource)
                global_provider = trace.get_tracer_provider()
                if set_global_tracer_provider and isinstance(
                    global_provider, trace.ProxyTracerProvider
                ):
                    trace.set_tracer_provider(lmnr_provider)

                obj._tracer_provider = lmnr_provider

                obj._tracer_provider.add_span_processor(obj._span_processor)

                # This is not a real instrumentation and does not generate telemetry
                # data, but it is required to ensure that OpenTelemetry context
                # propagation is enabled.
                # See the README at:
                # https://pypi.org/project/opentelemetry-instrumentation-threading/
                ThreadingInstrumentor().instrument()

                init_instrumentations(
                    tracer_provider=obj._tracer_provider,
                    instruments=instruments,
                    block_instruments=block_instruments,
                    client=obj._client,
                    async_client=obj._async_client,
                )

                cls.instance = obj

                # Force flushes for debug environments (e.g. local development)
                atexit.register(obj.exit_handler)

            return cls.instance

    def exit_handler(self):
        if isinstance(self._span_processor, LaminarSpanProcessor):
            self._span_processor.clear()
        self.flush()

    def _initialize_logger(self):
        self._logger = logging.getLogger(__name__)
        console_log_handler = logging.StreamHandler()
        console_log_handler.setFormatter(VerboseColorfulFormatter())
        self._logger.addHandler(console_log_handler)

    @staticmethod
    def set_static_params(
        resource_attributes: dict,
        enable_content_tracing: bool,
    ) -> None:
        TracerWrapper.resource_attributes = resource_attributes
        TracerWrapper.enable_content_tracing = enable_content_tracing

    @classmethod
    def verify_initialized(cls) -> bool:
        with cls._lock:
            return hasattr(cls, "instance") and hasattr(cls.instance, "_span_processor")

    @classmethod
    def clear(cls):
        if not cls.verify_initialized():
            return
        # Any state cleanup. Now used in between tests
        if isinstance(cls.instance._span_processor, LaminarSpanProcessor):
            cls.instance._span_processor.clear()

    def shutdown(self):
        if self._tracer_provider is None:
            return
        self._tracer_provider.shutdown()

    def flush(self):
        if not hasattr(self, "_span_processor"):
            self._logger.warning("TracerWrapper not fully initialized, cannot flush")
            return False
        return self._span_processor.force_flush()

    def get_tracer(self):
        if self._tracer_provider is None:
            return trace.get_tracer_provider().get_tracer(TRACER_NAME)
        return self._tracer_provider.get_tracer(TRACER_NAME)



import uuid

from opentelemetry.sdk.trace.export import (
    SpanProcessor,
    SpanExporter,
    BatchSpanProcessor,
    SimpleSpanProcessor,
)
from opentelemetry.sdk.trace import Span
from opentelemetry.context import Context, get_value, get_current, set_value

from lmnr.opentelemetry_lib.tracing.attributes import (
    SPAN_IDS_PATH,
    SPAN_INSTRUMENTATION_SOURCE,
    SPAN_LANGUAGE_VERSION,
    SPAN_PATH,
    SPAN_SDK_VERSION,
)
from lmnr.opentelemetry_lib.tracing.exporter import LaminarSpanExporter
from lmnr.opentelemetry_lib.tracing.context_properties import (
    _set_association_properties_attributes,
)
from lmnr.version import PYTHON_VERSION, __version__


class LaminarSpanProcessor(SpanProcessor):
    instance: BatchSpanProcessor | SimpleSpanProcessor
    __span_id_to_path: dict[int, list[str]] = {}
    __span_id_lists: dict[int, list[str]] = {}

    def __init__(
        self,
        base_url: str | None = None,
        port: int | None = None,
        api_key: str | None = None,
        timeout_seconds: int = 30,
        force_http: bool = False,
        max_export_batch_size: int = 512,
        disable_batch: bool = False,
        exporter: SpanExporter | None = None,
    ):
        self.exporter = exporter or LaminarSpanExporter(
            base_url=base_url,
            port=port,
            api_key=api_key,
            timeout_seconds=timeout_seconds,
            force_http=force_http,
        )
        self.instance = (
            SimpleSpanProcessor(self.exporter)
            if disable_batch
            else BatchSpanProcessor(
                self.exporter, max_export_batch_size=max_export_batch_size
            )
        )

    def on_start(self, span: Span, parent_context: Context | None = None):
        span_path_in_context = get_value("span_path", parent_context or get_current())
        parent_span_path = span_path_in_context or (
            self.__span_id_to_path.get(span.parent.span_id) if span.parent else None
        )
        parent_span_ids_path = (
            self.__span_id_lists.get(span.parent.span_id, []) if span.parent else []
        )
        span_path = parent_span_path + [span.name] if parent_span_path else [span.name]
        span_ids_path = parent_span_ids_path + [
            str(uuid.UUID(int=span.get_span_context().span_id))
        ]
        span.set_attribute(SPAN_PATH, span_path)
        span.set_attribute(SPAN_IDS_PATH, span_ids_path)
        set_value("span_path", span_path, get_current())
        self.__span_id_to_path[span.get_span_context().span_id] = span_path
        self.__span_id_lists[span.get_span_context().span_id] = span_ids_path

        span.set_attribute(SPAN_INSTRUMENTATION_SOURCE, "python")
        span.set_attribute(SPAN_SDK_VERSION, __version__)
        span.set_attribute(SPAN_LANGUAGE_VERSION, f"python@{PYTHON_VERSION}")

        association_properties = get_value("association_properties")
        if association_properties is not None:
            _set_association_properties_attributes(span, association_properties)
        self.instance.on_start(span, parent_context)

    def on_end(self, span: Span):
        self.instance.on_end(span)

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return self.instance.force_flush(timeout_millis)

    def shutdown(self):
        self.instance.shutdown()

    def clear(self):
        self.__span_id_to_path = {}
        self.__span_id_lists = {}



import logging
import traceback

from .config import (
    Config,
)
from google.genai import types
from google.genai._common import BaseModel
import pydantic
from opentelemetry.trace import Span
from typing import Any, Literal


class ToolCall(pydantic.BaseModel):
    name: str | None = pydantic.Field(default=None)
    id: str | None = pydantic.Field(default=None)
    arguments: dict[str, Any] = pydantic.Field(default={})


class ImageUrlInner(pydantic.BaseModel):
    url: str = pydantic.Field(default="")


class ImageUrl(pydantic.BaseModel):
    type: Literal["image_url"] = pydantic.Field(default="image_url")
    image_url: ImageUrlInner = pydantic.Field(default=ImageUrlInner())


class ProcessedContentPart(pydantic.BaseModel):
    content: str | None = pydantic.Field(default=None)
    function_call: ToolCall | None = pydantic.Field(default=None)
    image_url: ImageUrl | None = pydantic.Field(default=None)


def set_span_attribute(span: Span, name: str, value: str):
    if value is not None:
        if value != "":
            span.set_attribute(name, value)
    return


def dont_throw(func):
    """
    A decorator that wraps the passed in function and logs exceptions instead of throwing them.

    @param func: The function to wrap
    @return: The wrapper function
    """
    # Obtain a logger specific to the function's module
    logger = logging.getLogger(func.__module__)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.debug(
                "Laminar failed to trace in %s, error: %s",
                func.__name__,
                traceback.format_exc(),
            )
            if Config.exception_logger:
                Config.exception_logger(e)

    return wrapper


def to_dict(obj: BaseModel | pydantic.BaseModel | dict) -> dict[str, Any]:
    try:
        if isinstance(obj, BaseModel):
            return obj.model_dump()
        elif isinstance(obj, pydantic.BaseModel):
            return obj.model_dump()
        elif isinstance(obj, dict):
            return obj
        else:
            return dict(obj)
    except Exception:
        return dict(obj)


def get_content(
    content: (
        ProcessedContentPart | dict | list[ProcessedContentPart | dict] | str | None
    ),
) -> list[Any] | None:
    if isinstance(content, dict):
        return content.get("content") or content.get("image_url")
    if isinstance(content, ProcessedContentPart):
        if content.content and isinstance(content.content, str):
            return {
                "type": "text",
                "text": content.content,
            }
        elif content.image_url:
            return content.image_url.model_dump()
        else:
            return None
    elif isinstance(content, list):
        return [get_content(item) or "" for item in content if item is not None]
    elif isinstance(content, str):
        return {
            "type": "text",
            "text": content,
        }
    else:
        return None


def process_content_union(
    content: types.ContentUnion | types.ContentUnionDict,
    trace_id: str | None = None,
    span_id: str | None = None,
    message_index: int = 0,
) -> ProcessedContentPart | dict | list[ProcessedContentPart | dict] | None:
    if isinstance(content, types.Content):
        parts = to_dict(content).get("parts", [])
        return [_process_part(part) for part in parts]
    elif isinstance(content, list):
        return [_process_part_union(item) for item in content]
    elif isinstance(content, (types.Part, types.File, str)):
        return _process_part_union(content)
    elif isinstance(content, dict):
        if "parts" in content:
            return [
                _process_part_union(
                    item, trace_id, span_id, message_index, content_index
                )
                for content_index, item in enumerate(content.get("parts", []))
            ]
        else:
            # Assume it's PartDict
            return _process_part_union(content, trace_id, span_id, message_index)
    else:
        return None


def _process_part_union(
    content: types.PartDict | types.File | types.Part | str,
    trace_id: str | None = None,
    span_id: str | None = None,
    message_index: int = 0,
    content_index: int = 0,
) -> ProcessedContentPart | dict | None:
    if isinstance(content, str):
        return ProcessedContentPart(content=content)
    elif isinstance(content, types.File):
        content_dict = to_dict(content)
        name = (
            content_dict.get("name")
            or content_dict.get("display_name")
            or content_dict.get("uri")
        )
        return ProcessedContentPart(content=f"files/{name}")
    elif isinstance(content, (types.Part, dict)):
        return _process_part(content, trace_id, span_id, message_index, content_index)
    else:
        return None


def _process_part(
    content: types.Part,
    trace_id: str | None = None,
    span_id: str | None = None,
    message_index: int = 0,
    content_index: int = 0,
) -> ProcessedContentPart | dict | None:
    part_dict = to_dict(content)
    if part_dict.get("inline_data"):
        blob = to_dict(part_dict.get("inline_data"))
        if blob.get("mime_type").startswith("image/"):
            return _process_image_item(
                blob, trace_id, span_id, message_index, content_index
            )
        else:
            # currently, only images are supported
            return ProcessedContentPart(
                content=blob.get("mime_type") or "unknown_media"
            )
    elif part_dict.get("function_call"):
        return ProcessedContentPart(
            function_call=ToolCall(
                name=part_dict.get("function_call").get("name"),
                id=part_dict.get("function_call").get("id"),
                arguments=part_dict.get("function_call").get("args", {}),
            )
        )
    elif part_dict.get("text") is not None:
        return ProcessedContentPart(content=part_dict.get("text"))
    else:
        return None


def role_from_content_union(
    content: types.ContentUnion | types.ContentUnionDict,
) -> str | None:
    role = None
    if isinstance(content, types.Content):
        role = to_dict(content).get("role")
    elif isinstance(content, list) and len(content) > 0:
        role = role_from_content_union(content[0])
    elif isinstance(content, dict):
        role = content.get("role")
    else:
        return None
    return role
    # return "assistant" if role == "model" else role


def with_tracer_wrapper(func):
    """Helper for providing tracer for wrapper functions."""

    def _with_tracer(tracer, to_wrap):
        def wrapper(wrapped, instance, args, kwargs):
            return func(tracer, to_wrap, wrapped, instance, args, kwargs)

        return wrapper

    return _with_tracer


def _process_image_item(
    blob: dict[str, Any],
    trace_id: str,
    span_id: str,
    message_index: int,
    content_index: int,
) -> ProcessedContentPart | dict | None:
    # Convert to openai format, so backends can handle it
    return (
        ProcessedContentPart(
            image_url=ImageUrl(
                image_url=ImageUrlInner(
                    url=f"data:image/{blob.get('mime_type').split('/')[1]};base64,{blob.get('data')}",
                )
            )
        )
        if Config.convert_image_to_openai_format
        else blob
    )



from argparse import ArgumentParser
import asyncio
import glob
import importlib.util
import json
import os
import re
import sys
import urllib.request
import urllib.error
from pathlib import Path

from lmnr.sdk.evaluations import Evaluation

from .sdk.eval_control import PREPARE_ONLY, EVALUATION_INSTANCES
from .sdk.log import get_default_logger

LOG = get_default_logger(__name__)


EVAL_DIR = "evals"


def add_cursor_rules():
    """Download laminar.mdc file from a hardcoded public URL and save it to .cursor/rules/laminar.mdc"""
    # Hardcoded URL for the laminar.mdc file
    url = "https://raw.githubusercontent.com/lmnr-ai/lmnr/dev/rules/laminar.mdc"

    # Create .cursor/rules directory if it doesn't exist
    rules_dir = Path(".cursor/rules")
    rules_dir.mkdir(parents=True, exist_ok=True)

    # Define the target file path
    target_file = rules_dir / "laminar.mdc"

    try:
        LOG.info(f"Downloading laminar.mdc from {url}")

        # Download the file
        with urllib.request.urlopen(url) as response:
            content = response.read()

        # Write the content to the target file (this will overwrite if it exists)
        with open(target_file, "wb") as f:
            f.write(content)

        LOG.info(f"Successfully downloaded laminar.mdc to {target_file}")

    except urllib.error.URLError as e:
        LOG.error(f"Failed to download file from {url}: {e}")
        sys.exit(1)
    except Exception as e:
        LOG.error(f"Unexpected error: {e}")
        sys.exit(1)


async def run_evaluation(args):
    sys.path.append(os.getcwd())

    if len(args.file) == 0:
        files = [
            os.path.join(EVAL_DIR, f)
            for f in os.listdir(EVAL_DIR)
            if re.match(r".*_eval\.py$", f) or re.match(r"eval_.*\.py$", f)
        ]
        if len(files) == 0:
            LOG.error("No evaluation files found in `evals` directory")
            LOG.info(
                "Eval files must be located in the `evals` directory and must be named *_eval.py or eval_*.py"
            )
            return
        files.sort()
        LOG.info(f"Located {len(files)} evaluation files in {EVAL_DIR}")

    else:
        files = []
        for pattern in args.file:
            matches = glob.glob(pattern)
            if matches:
                files.extend(matches)
            else:
                # If no matches found, treat as literal filename
                files.append(pattern)

    prep_token = PREPARE_ONLY.set(True)
    scores = []
    try:
        for file in files:
            LOG.info(f"Running evaluation from {file}")
            file = os.path.abspath(file)
            name = "user_module" + file

            spec = importlib.util.spec_from_file_location(name, file)
            if spec is None or spec.loader is None:
                LOG.error(f"Could not load module specification from {file}")
                if args.continue_on_error:
                    continue
                return
            mod = importlib.util.module_from_spec(spec)
            sys.modules[name] = mod

            spec.loader.exec_module(mod)
            evaluations = []
            try:
                evaluations: list[Evaluation] | None = EVALUATION_INSTANCES.get()
                if evaluations is None:
                    raise LookupError()
            # may be raised by `get()` or manually by us above
            except LookupError:
                log_evaluation_instance_not_found()
                if args.continue_on_error:
                    continue
                return

            LOG.info(f"Loaded {len(evaluations)} evaluations from {file}")

            for evaluation in evaluations:
                try:
                    eval_scores = await evaluation.run()
                    scores.append(
                        {
                            "file": file,
                            "scores": eval_scores,
                        }
                    )
                except Exception as e:
                    LOG.error(f"Error running evaluation: {e}")
                    if not args.continue_on_error:
                        raise

        if args.output_file:
            with open(args.output_file, "w") as f:
                json.dump(scores, f, indent=2)
    finally:
        PREPARE_ONLY.reset(prep_token)


def log_evaluation_instance_not_found():
    LOG.warning(
        "Evaluation instance not found. "
        "`evaluate` must be called at the top level of the file, "
        "not inside a function when running evaluations from the CLI."
    )


def cli():
    parser = ArgumentParser(
        prog="lmnr",
        description="CLI for Laminar. Call `lmnr [subcommand] --help` for more information on each subcommand.",
    )

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    parser_eval = subparsers.add_parser(
        "eval",
        description="Run an evaluation",
        help="Run an evaluation",
    )
    parser_eval.add_argument(
        "file",
        nargs="*",
        help="Files or a file containing the evaluation to run."
        + "If no file name is provided, all evaluation files in the `evals` directory are run as long"
        + "as they match *_eval.py or eval_*.py",
        default=[],
    )

    parser_eval.add_argument(
        "--continue-on-error",
        action="store_true",
        default=False,
        help="Continue execution upon errors",
    )

    parser_eval.add_argument(
        "--output-file",
        help="Output file to write the results to. Outputs are written in JSON format.",
        nargs="?",
    )

    subparsers.add_parser(
        "add-cursor-rules",
        description="Download laminar.mdc file and add it to .cursor/rules",
        help="Download laminar.mdc file and add it to .cursor/rules",
    )

    parsed = parser.parse_args()
    if parsed.subcommand == "eval":
        asyncio.run(run_evaluation(parsed))
    elif parsed.subcommand == "add-cursor-rules":
        add_cursor_rules()
    else:
        parser.print_help()



from lmnr.opentelemetry_lib.decorators import (
    entity_method,
    aentity_method,
    json_dumps,
)
from opentelemetry.trace import INVALID_SPAN, get_current_span

from typing import Any, Callable, Literal, TypeVar, cast
from typing_extensions import ParamSpec

from lmnr.opentelemetry_lib.tracing.attributes import SESSION_ID
from lmnr.sdk.log import get_default_logger

from .utils import is_async

logger = get_default_logger(__name__)

P = ParamSpec("P")
R = TypeVar("R")


def observe(
    *,
    name: str | None = None,
    session_id: str | None = None,
    user_id: str | None = None,
    ignore_input: bool = False,
    ignore_output: bool = False,
    span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
    ignore_inputs: list[str] | None = None,
    metadata: dict[str, Any] | None = None,
    tags: list[str] | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """The main decorator entrypoint for Laminar. This is used to wrap
    functions and methods to create spans.

    Args:
        name (str | None, optional): Name of the span. Function name is used if\
            not specified. Defaults to None.
        session_id (str | None, optional): Session ID to associate with the\
            span and the following context. Defaults to None.
        user_id (str | None, optional): User ID to associate with the span and\
            the following context. This is different from ID of a Laminar user.
            Defaults to None.
        ignore_input (bool, optional): Whether to ignore ALL input of the\
            wrapped function. Defaults to False.
        ignore_output (bool, optional): Whether to ignore ALL output of the\
            wrapped function. Defaults to False.
        span_type (Literal["DEFAULT", "LLM", "TOOL"], optional): Type of the span.
            Defaults to "DEFAULT".
        ignore_inputs (list[str] | None, optional): List of input keys to\
            ignore. For example, if the wrapped function takes three arguments\
            def foo(a, b, `sensitive_data`), and you want to ignore the\
            `sensitive_data` argument, you can pass ["sensitive_data"] to\
            this argument. Defaults to None.
        metadata (dict[str, Any] | None, optional): Metadata to associate with\
            the trace. Must be JSON serializable. Defaults to None.
        tags (list[str] | None, optional): Tags to associate with the trace.
            Defaults to None.
    Raises:
        Exception: re-raises the exception if the wrapped function raises an\
            exception

    Returns:
        R: Returns the result of the wrapped function
    """

    def decorator(func: Callable) -> Callable:
        current_span = get_current_span()
        if current_span != INVALID_SPAN:
            if session_id is not None:
                current_span.set_attribute(SESSION_ID, session_id)
        association_properties = {}
        if session_id is not None:
            association_properties["session_id"] = session_id
        if user_id is not None:
            association_properties["user_id"] = user_id
        if metadata is not None:
            association_properties.update(
                {
                    f"metadata.{k}": (
                        v if isinstance(v, (str, int, float, bool)) else json_dumps(v)
                    )
                    for k, v in metadata.items()
                }
            )
        if tags is not None:
            if not isinstance(tags, list) or not all(
                isinstance(tag, str) for tag in tags
            ):
                logger.warning("Tags must be a list of strings. Tags will be ignored.")
            else:
                association_properties["tags"] = tags
        result = (
            aentity_method(
                name=name,
                ignore_input=ignore_input,
                ignore_output=ignore_output,
                span_type=span_type,
                ignore_inputs=ignore_inputs,
                association_properties=association_properties,
            )(func)
            if is_async(func)
            else entity_method(
                name=name,
                ignore_input=ignore_input,
                ignore_output=ignore_output,
                span_type=span_type,
                ignore_inputs=ignore_inputs,
                association_properties=association_properties,
            )(func)
        )
        return result

    return cast(Callable, decorator)



from typing import Callable, Coroutine


class Config:
    exception_logger = None
    upload_base64_image: (
        Callable[[str, str, str, str], Coroutine[None, None, str]] | None
    ) = None
    convert_image_to_openai_format: bool = True



"""OpenTelemetry Google Generative AI API instrumentation"""

from collections import defaultdict
import json
import logging
import os
from typing import AsyncGenerator, Callable, Collection, Generator

from google.genai import types

from .config import (
    Config,
)
from .utils import (
    ProcessedContentPart,
    dont_throw,
    get_content,
    role_from_content_union,
    set_span_attribute,
    process_content_union,
    to_dict,
    with_tracer_wrapper,
)
from opentelemetry.trace import Tracer
from wrapt import wrap_function_wrapper

from opentelemetry import context as context_api
from opentelemetry.trace import get_tracer, SpanKind, Span
from opentelemetry.semconv._incubating.attributes import gen_ai_attributes

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY, unwrap

from opentelemetry.semconv_ai import (
    SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY,
    SpanAttributes,
    LLMRequestTypeValues,
)

logger = logging.getLogger(__name__)

_instruments = ("google-genai >= 1.0.0",)

WRAPPED_METHODS = [
    {
        "package": "google.genai.models",
        "object": "Models",
        "method": "generate_content",
        "span_name": "gemini.generate_content",
        "is_streaming": False,
        "is_async": False,
    },
    {
        "package": "google.genai.models",
        "object": "AsyncModels",
        "method": "generate_content",
        "span_name": "gemini.generate_content",
        "is_streaming": False,
        "is_async": True,
    },
    {
        "package": "google.genai.models",
        "object": "Models",
        "method": "generate_content_stream",
        "span_name": "gemini.generate_content_stream",
        "is_streaming": True,
        "is_async": False,
    },
    {
        "package": "google.genai.models",
        "object": "AsyncModels",
        "method": "generate_content_stream",
        "span_name": "gemini.generate_content_stream",
        "is_streaming": True,
        "is_async": True,
    },
]


def should_send_prompts():
    return (
        os.getenv("TRACELOOP_TRACE_CONTENT") or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


@dont_throw
def _set_request_attributes(span, args, kwargs):
    config_dict = to_dict(kwargs.get("config", {}))
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_MODEL, kwargs.get("model")
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_TEMPERATURE,
        config_dict.get("temperature"),
    )
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_TOP_P, config_dict.get("top_p")
    )
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_TOP_K, config_dict.get("top_k")
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_CHOICE_COUNT,
        config_dict.get("candidate_count"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_MAX_TOKENS,
        config_dict.get("max_output_tokens"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_STOP_SEQUENCES,
        config_dict.get("stop_sequences"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_FREQUENCY_PENALTY,
        config_dict.get("frequency_penalty"),
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_REQUEST_PRESENCE_PENALTY,
        config_dict.get("presence_penalty"),
    )
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_REQUEST_SEED, config_dict.get("seed")
    )

    tools: list[types.FunctionDeclaration] = []
    if kwargs.get("tools"):
        for tool in kwargs.get("tools"):
            if isinstance(tool, types.Tool):
                tools += tool.function_declarations or []
            elif isinstance(tool, Callable):
                tools.append(types.FunctionDeclaration.from_callable(tool))
    for tool_num, tool in enumerate(tools):
        set_span_attribute(
            span,
            f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{tool_num}.name",
            to_dict(tool).get("name"),
        )
        set_span_attribute(
            span,
            f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{tool_num}.description",
            to_dict(tool).get("description"),
        )
        set_span_attribute(
            span,
            f"{SpanAttributes.LLM_REQUEST_FUNCTIONS}.{tool_num}.parameters",
            to_dict(tool).get("parameters"),
        )

    if should_send_prompts():
        i = 0
        system_instruction: types.ContentUnion | None = config_dict.get(
            "system_instruction"
        )
        if system_instruction:
            set_span_attribute(
                span,
                f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.content",
                (get_content(process_content_union(system_instruction)) or {}).get("text", ""),
            )
            set_span_attribute(
                span, f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.role", "system"
            )
            i += 1
        contents = kwargs.get("contents", [])
        if not isinstance(contents, list):
            contents = [contents]
        for content in contents:
            processed_content = process_content_union(content)
            content_str = get_content(processed_content)
            set_span_attribute(
                span,
                f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.content",
                (
                    content_str
                    if isinstance(content_str, str)
                    else json.dumps(content_str)
                ),
            )
            blocks = (
                processed_content
                if isinstance(processed_content, list)
                else [processed_content]
            )
            for j, block in enumerate(blocks):
                block_dict = to_dict(block)
                if not block_dict.get("function_call"):
                    continue
                function_call = to_dict(block_dict.get("function_call", {}))
                set_span_attribute(
                    span,
                    f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.tool_calls.{j}.name",
                    function_call.get("name"),
                )
                set_span_attribute(
                    span,
                    f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.tool_calls.{j}.id",
                    function_call.get("id"),
                )
                set_span_attribute(
                    span,
                    f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.tool_calls.{j}.arguments",
                    json.dumps(function_call.get("arguments")),
                )
            set_span_attribute(
                span,
                f"{gen_ai_attributes.GEN_AI_PROMPT}.{i}.role",
                role_from_content_union(content) or "user",
            )
            i += 1


@dont_throw
def _set_response_attributes(span, response: types.GenerateContentResponse):
    candidates = response.candidates or []
    set_span_attribute(
        span, gen_ai_attributes.GEN_AI_RESPONSE_ID, to_dict(response).get("response_id")
    )
    set_span_attribute(
        span,
        gen_ai_attributes.GEN_AI_RESPONSE_MODEL,
        to_dict(response).get("model_version"),
    )

    if response.usage_metadata:
        usage_dict = to_dict(response.usage_metadata)
        set_span_attribute(
            span,
            gen_ai_attributes.GEN_AI_USAGE_INPUT_TOKENS,
            usage_dict.get("prompt_token_count"),
        )
        set_span_attribute(
            span,
            gen_ai_attributes.GEN_AI_USAGE_OUTPUT_TOKENS,
            usage_dict.get("candidates_token_count"),
        )
        set_span_attribute(
            span,
            SpanAttributes.LLM_USAGE_TOTAL_TOKENS,
            usage_dict.get("total_token_count"),
        )
        set_span_attribute(
            span,
            SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS,
            usage_dict.get("cached_content_token_count"),
        )

    if should_send_prompts():
        set_span_attribute(
            span, f"{gen_ai_attributes.GEN_AI_COMPLETION}.0.role", "model"
        )
        candidates_list = candidates if isinstance(candidates, list) else [candidates]
        for i, candidate in enumerate(candidates_list):
            processed_content = process_content_union(candidate.content)
            if isinstance(processed_content, list):
                if all(
                    isinstance(item, dict) and item.get("type") == "text"
                    for item in processed_content
                ):
                    content_str = processed_content[0]["text"]
                elif all(
                    isinstance(item, ProcessedContentPart) and item.content
                    for item in processed_content
                ):
                    content_str = processed_content[0].content
                else:
                    content_str = get_content(processed_content)
            else:
                content_str = get_content(processed_content)
            set_span_attribute(
                span, f"{gen_ai_attributes.GEN_AI_COMPLETION}.{i}.role", "model"
            )
            set_span_attribute(
                span,
                f"{gen_ai_attributes.GEN_AI_COMPLETION}.{i}.content",
                (
                    content_str
                    if isinstance(content_str, str)
                    else json.dumps(content_str)
                ),
            )
            blocks = (
                processed_content
                if isinstance(processed_content, list)
                else [processed_content]
            )
            for j, block in enumerate(blocks):
                block_dict = to_dict(block)
                if not block_dict.get("function_call"):
                    continue
                function_call = to_dict(block_dict.get("function_call", {}))
                set_span_attribute(
                    span,
                    f"{gen_ai_attributes.GEN_AI_COMPLETION}.{i}.tool_calls.{j}.name",
                    function_call.get("name"),
                )
                set_span_attribute(
                    span,
                    f"{gen_ai_attributes.GEN_AI_COMPLETION}.{i}.tool_calls.{j}.id",
                    function_call.get("id"),
                )
                set_span_attribute(
                    span,
                    f"{gen_ai_attributes.GEN_AI_COMPLETION}.{i}.tool_calls.{j}.arguments",
                    json.dumps(function_call.get("arguments")),
                )


@dont_throw
def _build_from_streaming_response(
    span: Span, response: Generator[types.GenerateContentResponse, None, None]
) -> Generator[types.GenerateContentResponse, None, None]:
    final_parts = []
    role = "model"
    aggregated_usage_metadata = defaultdict(int)
    model_version = None
    for chunk in response:
        if chunk.model_version:
            model_version = chunk.model_version

        if chunk.candidates:
            # Currently gemini throws an error if you pass more than one candidate
            # with streaming
            if chunk.candidates and len(chunk.candidates) > 0:
                final_parts += chunk.candidates[0].content.parts or []
                role = chunk.candidates[0].content.role or role
        if chunk.usage_metadata:
            usage_dict = to_dict(chunk.usage_metadata)
            # prompt token count is sent in every chunk
            # (and is less by 1 in the last chunk, so we set it once);
            # total token count in every chunk is greater by prompt token count than it should be,
            # thus this awkward logic here
            if aggregated_usage_metadata.get("prompt_token_count") is None:
                aggregated_usage_metadata["prompt_token_count"] = (
                    usage_dict.get("prompt_token_count") or 0
                )
                aggregated_usage_metadata["total_token_count"] = (
                    usage_dict.get("total_token_count") or 0
                )
            aggregated_usage_metadata["candidates_token_count"] += (
                usage_dict.get("candidates_token_count") or 0
            )
            aggregated_usage_metadata["total_token_count"] += (
                usage_dict.get("candidates_token_count") or 0
            )
        yield chunk

    compound_response = types.GenerateContentResponse(
        candidates=[
            {
                "content": {
                    "parts": final_parts,
                    "role": role,
                },
            }
        ],
        usage_metadata=types.GenerateContentResponseUsageMetadataDict(
            **aggregated_usage_metadata
        ),
        model_version=model_version,
    )
    if span.is_recording():
        _set_response_attributes(span, compound_response)
    span.end()


@dont_throw
async def _abuild_from_streaming_response(
    span: Span, response: AsyncGenerator[types.GenerateContentResponse, None]
) -> AsyncGenerator[types.GenerateContentResponse, None]:
    final_parts = []
    role = "model"
    aggregated_usage_metadata = defaultdict(int)
    model_version = None
    async for chunk in response:
        if chunk.candidates:
            # Currently gemini throws an error if you pass more than one candidate
            # with streaming
            if chunk.candidates and len(chunk.candidates) > 0:
                final_parts += chunk.candidates[0].content.parts or []
                role = chunk.candidates[0].content.role or role
                if chunk.model_version:
                    model_version = chunk.model_version
        if chunk.usage_metadata:
            usage_dict = to_dict(chunk.usage_metadata)
            # prompt token count is sent in every chunk
            # (and is less by 1 in the last chunk, so we set it once);
            # total token count in every chunk is greater by prompt token count than it should be,
            # thus this awkward logic here
            if aggregated_usage_metadata.get("prompt_token_count") is None:
                aggregated_usage_metadata["prompt_token_count"] = usage_dict.get(
                    "prompt_token_count"
                )
                aggregated_usage_metadata["total_token_count"] = usage_dict.get(
                    "total_token_count"
                )
            aggregated_usage_metadata["candidates_token_count"] += (
                usage_dict.get("candidates_token_count") or 0
            )
            aggregated_usage_metadata["total_token_count"] += (
                usage_dict.get("candidates_token_count") or 0
            )
        yield chunk

    compound_response = types.GenerateContentResponse(
        candidates=[
            {
                "content": {
                    "parts": final_parts,
                    "role": role,
                },
            }
        ],
        usage_metadata=types.GenerateContentResponseUsageMetadataDict(
            **aggregated_usage_metadata
        ),
        model_version=model_version,
    )
    if span.is_recording():
        _set_response_attributes(span, compound_response)
    span.end()


@with_tracer_wrapper
def _wrap(tracer: Tracer, to_wrap, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return wrapped(*args, **kwargs)

    span = tracer.start_span(
        to_wrap.get("span_name"),
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.LLM_SYSTEM: "gemini",
            SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.COMPLETION.value,
        },
    )

    if span.is_recording():
        _set_request_attributes(span, args, kwargs)

    if to_wrap.get("is_streaming"):
        return _build_from_streaming_response(span, wrapped(*args, **kwargs))
    else:
        response = wrapped(*args, **kwargs)

    if span.is_recording():
        _set_response_attributes(span, response)

    span.end()
    return response


@with_tracer_wrapper
async def _awrap(tracer: Tracer, to_wrap, wrapped, instance, args, kwargs):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return await wrapped(*args, **kwargs)

    span = tracer.start_span(
        to_wrap.get("span_name"),
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.LLM_SYSTEM: "gemini",
            SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.COMPLETION.value,
        },
    )

    if span.is_recording():
        _set_request_attributes(span, args, kwargs)

    if to_wrap.get("is_streaming"):
        return _abuild_from_streaming_response(span, await wrapped(*args, **kwargs))
    else:
        response = await wrapped(*args, **kwargs)

    if span.is_recording():
        _set_response_attributes(span, response)

    span.end()
    return response


class GoogleGenAiSdkInstrumentor(BaseInstrumentor):
    """An instrumentor for Google GenAI's client library."""

    def __init__(
        self,
        exception_logger=None,
        upload_base64_image=None,
        convert_image_to_openai_format=True,
    ):
        super().__init__()
        Config.exception_logger = exception_logger
        Config.upload_base64_image = upload_base64_image
        Config.convert_image_to_openai_format = convert_image_to_openai_format

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "0.0.1a1", tracer_provider)

        for wrapped_method in WRAPPED_METHODS:
            wrap_function_wrapper(
                wrapped_method.get("package"),
                f"{wrapped_method.get('object')}.{wrapped_method.get('method')}",
                (
                    _awrap(tracer, wrapped_method)
                    if wrapped_method.get("is_async")
                    else _wrap(tracer, wrapped_method)
                ),
            )

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            unwrap(
                f"{wrapped_method.get('package')}.{wrapped_method.get('object')}",
                wrapped_method.get("method"),
            )



import logging
import uuid

from lmnr.opentelemetry_lib.utils.package_check import is_package_installed
from lmnr.sdk.browser.pw_utils import handle_navigation_async, handle_navigation_sync
from lmnr.sdk.browser.utils import with_tracer_and_client_wrapper
from lmnr.sdk.client.asynchronous.async_client import AsyncLaminarClient
from lmnr.sdk.client.synchronous.sync_client import LaminarClient
from lmnr.version import __version__

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.trace import (
    get_tracer,
    Tracer,
    get_current_span,
    Span,
    INVALID_SPAN,
    set_span_in_context,
)
from opentelemetry.context import get_current
from typing import Collection
from wrapt import wrap_function_wrapper

try:
    if is_package_installed("playwright"):
        from playwright.async_api import Browser, BrowserContext, Page
        from playwright.sync_api import (
            Browser as SyncBrowser,
            BrowserContext as SyncBrowserContext,
            Page as SyncPage,
        )
    elif is_package_installed("patchright"):
        from patchright.async_api import Browser, BrowserContext, Page
        from patchright.sync_api import (
            Browser as SyncBrowser,
            BrowserContext as SyncBrowserContext,
            Page as SyncPage,
        )
    else:
        raise ImportError(
            "Attempted to import lmnr.sdk.browser.playwright_otel, but neither "
            "playwright nor patchright is installed. Use `pip install playwright` "
            "or `pip install patchright` to install one of the supported browsers."
        )
except ImportError as e:
    raise ImportError(
        f"Attempted to import {__file__}, but it is designed "
        "to patch Playwright, which is not installed. Use `pip install playwright` "
        "or `pip install patchright` to install Playwright or remove this import."
    ) from e

# all available versions at https://pypi.org/project/playwright/#history
_instruments = ("playwright >= 1.9.0",)
logger = logging.getLogger(__name__)

_context_spans: dict[str, Span] = {}


@with_tracer_and_client_wrapper
def _wrap_new_page(
    tracer: Tracer, client: LaminarClient, to_wrap, wrapped, instance, args, kwargs
):
    with tracer.start_as_current_span(
        f"{to_wrap.get('object')}.{to_wrap.get('method')}"
    ) as span:
        page = wrapped(*args, **kwargs)
        session_id = str(uuid.uuid4().hex)
        trace_id = format(get_current_span().get_span_context().trace_id, "032x")
        span.set_attribute("lmnr.internal.has_browser_session", True)
        handle_navigation_sync(page, session_id, trace_id, client)
        return page


@with_tracer_and_client_wrapper
async def _wrap_new_page_async(
    tracer: Tracer, client: AsyncLaminarClient, to_wrap, wrapped, instance, args, kwargs
):
    with tracer.start_as_current_span(
        f"{to_wrap.get('object')}.{to_wrap.get('method')}"
    ) as span:
        page = await wrapped(*args, **kwargs)
        session_id = str(uuid.uuid4().hex)
        trace_id = format(span.get_span_context().trace_id, "032x")
        span.set_attribute("lmnr.internal.has_browser_session", True)
        await handle_navigation_async(page, session_id, trace_id, client)
        return page


@with_tracer_and_client_wrapper
def _wrap_new_browser_sync(
    tracer: Tracer, client: LaminarClient, to_wrap, wrapped, instance, args, kwargs
):
    global _context_spans
    browser: SyncBrowser = wrapped(*args, **kwargs)
    session_id = str(uuid.uuid4().hex)
    for context in browser.contexts:
        span = get_current_span()
        if span == INVALID_SPAN:
            span = tracer.start_span(
                name=f"{to_wrap.get('object')}.{to_wrap.get('method')}"
            )
            set_span_in_context(span, get_current())
            _context_spans[id(context)] = span
        span.set_attribute("lmnr.internal.has_browser_session", True)
        trace_id = format(span.get_span_context().trace_id, "032x")

        def handle_page_navigation(page: SyncPage):
            return handle_navigation_sync(page, session_id, trace_id, client)

        context.on(
            "page",
            handle_page_navigation,
        )

        for page in context.pages:
            handle_navigation_sync(page, session_id, trace_id, client)
    return browser


@with_tracer_and_client_wrapper
async def _wrap_new_browser_async(
    tracer: Tracer, client: AsyncLaminarClient, to_wrap, wrapped, instance, args, kwargs
):
    global _context_spans
    browser: Browser = await wrapped(*args, **kwargs)
    session_id = str(uuid.uuid4().hex)
    for context in browser.contexts:
        span = get_current_span()
        if span == INVALID_SPAN:
            span = tracer.start_span(
                name=f"{to_wrap.get('object')}.{to_wrap.get('method')}"
            )
            set_span_in_context(span, get_current())
            _context_spans[id(context)] = span
        span.set_attribute("lmnr.internal.has_browser_session", True)
        trace_id = format(span.get_span_context().trace_id, "032x")

        async def handle_page_navigation(page: Page):
            return await handle_navigation_async(page, session_id, trace_id, client)

        context.on("page", handle_page_navigation)
        for page in context.pages:
            await handle_navigation_async(page, session_id, trace_id, client)
    return browser


@with_tracer_and_client_wrapper
def _wrap_new_context_sync(
    tracer: Tracer, client: LaminarClient, to_wrap, wrapped, instance, args, kwargs
):
    context: SyncBrowserContext = wrapped(*args, **kwargs)
    session_id = str(uuid.uuid4().hex)
    span = get_current_span()
    if span == INVALID_SPAN:
        span = tracer.start_span(
            name=f"{to_wrap.get('object')}.{to_wrap.get('method')}"
        )
        set_span_in_context(span, get_current())
        _context_spans[id(context)] = span
    span.set_attribute("lmnr.internal.has_browser_session", True)
    trace_id = format(span.get_span_context().trace_id, "032x")

    def handle_page_navigation(page: SyncPage):
        return handle_navigation_sync(page, session_id, trace_id, client)

    context.on(
        "page",
        handle_page_navigation,
    )
    for page in context.pages:
        handle_navigation_sync(page, session_id, trace_id, client)
    return context


@with_tracer_and_client_wrapper
async def _wrap_new_context_async(
    tracer: Tracer, client: AsyncLaminarClient, to_wrap, wrapped, instance, args, kwargs
):
    context: BrowserContext = await wrapped(*args, **kwargs)
    session_id = str(uuid.uuid4().hex)
    span = get_current_span()
    if span == INVALID_SPAN:
        span = tracer.start_span(
            name=f"{to_wrap.get('object')}.{to_wrap.get('method')}"
        )
        set_span_in_context(span, get_current())
        _context_spans[id(context)] = span
    span.set_attribute("lmnr.internal.has_browser_session", True)
    trace_id = format(span.get_span_context().trace_id, "032x")

    async def handle_page_navigation(page):
        return await handle_navigation_async(page, session_id, trace_id, client)

    context.on("page", handle_page_navigation)
    for page in context.pages:
        await handle_navigation_async(page, session_id, trace_id, client)
    return context


@with_tracer_and_client_wrapper
def _wrap_close_browser_sync(
    tracer: Tracer,
    client: LaminarClient,
    to_wrap,
    wrapped,
    instance: SyncBrowser,
    args,
    kwargs,
):
    global _context_spans
    for context in instance.contexts:
        key = id(context)
        span = _context_spans.get(key)
        if span:
            if span.is_recording():
                span.end()
            _context_spans.pop(key)
    return wrapped(*args, **kwargs)


@with_tracer_and_client_wrapper
async def _wrap_close_browser_async(
    tracer: Tracer,
    client: AsyncLaminarClient,
    to_wrap,
    wrapped,
    instance: Browser,
    args,
    kwargs,
):
    global _context_spans
    for context in instance.contexts:
        key = id(context)
        span = _context_spans.get(key)
        if span:
            if span.is_recording():
                span.end()
            _context_spans.pop(key)
    return await wrapped(*args, **kwargs)


WRAPPED_METHODS = [
    {
        "package": "playwright.sync_api",
        "object": "BrowserContext",
        "method": "new_page",
        "wrapper": _wrap_new_page,
    },
    {
        "package": "playwright.sync_api",
        "object": "Browser",
        "method": "new_page",
        "wrapper": _wrap_new_page,
    },
    {
        "package": "playwright.sync_api",
        "object": "BrowserType",
        "method": "launch",
        "wrapper": _wrap_new_browser_sync,
    },
    {
        "package": "playwright.sync_api",
        "object": "BrowserType",
        "method": "connect",
        "wrapper": _wrap_new_browser_sync,
    },
    {
        "package": "playwright.sync_api",
        "object": "BrowserType",
        "method": "connect_over_cdp",
        "wrapper": _wrap_new_browser_sync,
    },
    {
        "package": "playwright.sync_api",
        "object": "Browser",
        "method": "close",
        "wrapper": _wrap_close_browser_sync,
    },
    {
        "package": "playwright.sync_api",
        "object": "Browser",
        "method": "new_context",
        "wrapper": _wrap_new_context_sync,
    },
    {
        "package": "playwright.sync_api",
        "object": "BrowserType",
        "method": "launch_persistent_context",
        "wrapper": _wrap_new_context_sync,
    },
]

WRAPPED_METHODS_ASYNC = [
    {
        "package": "playwright.async_api",
        "object": "BrowserContext",
        "method": "new_page",
        "wrapper": _wrap_new_page_async,
    },
    {
        "package": "playwright.async_api",
        "object": "Browser",
        "method": "new_page",
        "wrapper": _wrap_new_page_async,
    },
    {
        "package": "playwright.async_api",
        "object": "BrowserType",
        "method": "launch",
        "wrapper": _wrap_new_browser_async,
    },
    {
        "package": "playwright.async_api",
        "object": "BrowserType",
        "method": "connect",
        "wrapper": _wrap_new_browser_async,
    },
    {
        "package": "playwright.async_api",
        "object": "BrowserType",
        "method": "connect_over_cdp",
        "wrapper": _wrap_new_browser_async,
    },
    {
        "package": "playwright.async_api",
        "object": "Browser",
        "method": "close",
        "wrapper": _wrap_close_browser_async,
    },
    {
        "package": "playwright.async_api",
        "object": "Browser",
        "method": "new_context",
        "wrapper": _wrap_new_context_async,
    },
    {
        "package": "playwright.async_api",
        "object": "BrowserType",
        "method": "launch_persistent_context",
        "wrapper": _wrap_new_context_async,
    },
]


class PlaywrightInstrumentor(BaseInstrumentor):
    def __init__(self, client: LaminarClient, async_client: AsyncLaminarClient):
        super().__init__()
        self.client = client
        self.async_client = async_client

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    wrapped_method.get("wrapper")(
                        tracer,
                        self.client,
                        wrapped_method,
                    ),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some module is missing

        # Wrap async methods
        for wrapped_method in WRAPPED_METHODS_ASYNC:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    wrapped_method.get("wrapper")(
                        tracer,
                        self.async_client,
                        wrapped_method,
                    ),
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some module is missing

    def _uninstrument(self, **kwargs):
        # Unwrap methods
        global _context_spans
        for wrapped_method in WRAPPED_METHODS + WRAPPED_METHODS_ASYNC:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            unwrap(wrap_package, f"{wrap_object}.{wrap_method}")
            for span in _context_spans.values():
                if span.is_recording():
                    span.end()
            _context_spans = {}
